import copy
import copy
from report_ranger.errors import InputError
import os
from pathlib import Path
import re
from jinja2 import Template
import jinja2
from jinja2.runtime import Macro
from jinja2.runtime import Macro
import yaml
import json
from report_ranger.imports.imports import import_csv, import_xlsx
import logging
from report_ranger.imports import section
from report_ranger.helpers import make_list
from report_ranger.imports import vulnerability
import traceback
import datetime
try:
    import tomllib
    hastomlib = True
except ModuleNotFoundError:
    # Error handling
    hastomlib = False
    pass

from report_ranger.utils.jinja_helpers import log_jinja2_error

log = logging.getLogger(__name__)

yamlmatch = re.compile(r'^[\.\=\-]{3}$')
jsonmatch = re.compile(r'^\;\;\;$')


def process_included_header(headers, includedheaders):
    """ Process the included headers using an overlay routine:
    - Don't overwrite values
    - Add new values if they're not there
    - Append to lists
    """
    if isinstance(includedheaders, dict):
        if not isinstance(headers, dict):
            return headers
        for header in includedheaders:
            if header in headers:
                header = process_included_header(
                    headers[header], includedheaders[header])
            else:
                headers[header] = includedheaders[header]
        return headers
    elif isinstance(includedheaders, list):
        if not isinstance(headers, dict):
            return headers
        headers.extend(includedheaders)
        return headers
    else:
        return headers


def process_template(headers, markdown, env=None, name='', filename=''):
    try:
        template = Template(markdown)  # Start jinja2
        env.push(headers)  # Add our new headers in
        output = template.render(env.get_env())
        env.pop()  # Take away the headers

    except jinja2.exceptions.TemplateSyntaxError as error:
        log.error(f"Error in processing {name} at {filename}")
        log.error(f"Message: {error.message}. Line number: {error.lineno}.")
        log_jinja2_error(markdown, error)
        raise Exception(f"Error reading {name}: {error.message} at lineno {error.lineno}")

    return output

def _get_resultant_path(file:str, filemapper:dict={}, pathlist:list=[]):
    # Try CWD
    filepath = Path(file)
    if filepath.exists():
        return filepath
    
    # Try includemapper
    if file in filemapper:
        file = filemapper[file]
        return Path(file)
    
    # Try pathlist
    if len(pathlist) > 0:
        for path in pathlist:
            filepath = Path(path) / Path(file)
            if Path(path) not in filepath.parents:
                log.warn(f"Path {filepath} seems to be going up in directories, skipping")
                raise Exception("Could not get resultant path")
            if filepath.exists():
                return filepath
    raise Exception("Could not find file")

def _process_includes(headers, env=None, includemapper={}, pathlist=[], watcher=None):
    for included in make_list(headers['include']):
        log.info(f"Including headers from file {included}")
        try:
            included = _get_resultant_path(included, includemapper, pathlist)
            includedheaders, _ = markdown_from_file(included, pathlist=pathlist, env=env, watcher=watcher)
            headers = process_included_header(headers, includedheaders)
        except Exception as e:
            log.warn(f"Could not include headers from file {included}: {e.args}")
    return headers

def _process_include_markdown(markdown, includemapper, pathlist, watcher=None):
    try:
        included = _get_resultant_path(markdown, includemapper, pathlist)
        log.debug(f"Processing included markdown")
        _, included_markdown = markdown_from_file(included, pathlist=pathlist, process_imports=False, process_includes=False, watcher=watcher)
        return included_markdown
    except Exception as e:
        log.warn(f"Could not include headers from file {included}: {e.args}")

def _check_import_directory(importdict):
    """ Run the standard checks for directory imports including vulnerabilities and sections
    """
    if 'variable' not in importdict or not isinstance(importdict['variable'], str):
        raise Exception("Import item does not have variable name")
    if not isinstance(importdict, dict):
        raise Exception(f"Import item '{importdict['variable']}' in front matter not a dict")
    if 'directory' not in importdict or not isinstance(importdict['directory'], str):
        raise Exception(f"Import item '{importdict['variable']}' does not have a directory")
    return True

def _process_imports(headers, env=None, watcher=None, includemapper:dict={}, pathlist:dict={}):
    new_headers = {}
    if not 'import' in headers:
        return {}
    if not isinstance(headers['import'], dict):
        log.warn("import front matter variable not a dict, skipping")
    else:
        if 'vulnerabilities' in headers['import']:
            vulnerabilities = make_list(headers['import']['vulnerabilities'])
            for vulnerabilitiesimport in vulnerabilities:
                if vulnerabilitiesimport['variable'] in headers:
                    log.warn(
                        f"Vulnerabilities import variable name {vulnerabilitiesimport['variable']} already in front matter, skipping")
                    continue
                
                try:
                    _check_import_directory(vulnerabilitiesimport)
                except Exception as e:
                    log.warn(f"Error importing vulnerabilities: {e.args}")

                vulndir = vulnerabilitiesimport['directory']
                log.info("Vulnerability directory included: {}".format(vulndir))

                vulnerabilitylist = vulnerability.VulnerabilityList()

                try:
                    # Get the vulnerability validation (if it exists) from the env
                    vulnerability_validation = {}
                    if 'validation' in vulnerabilitiesimport:
                        validation = vulnerabilitiesimport['validation']
                        if validation is dict and 'vulnerability' in validation:
                            vulnerability_validation = validation['vulnerability']

                    if env == None:
                        log.warn("Exception raising because env is None in _process_imports vulnerability import")
                        raise Exception("env is None")

                    vulnerabilitylist.add_from_dir(
                        vulndir, env, env.get("ra"), vulnerability_validation=vulnerability_validation, watcher=watcher)

                    if 'updated_date' in vulnerabilitiesimport:
                        updated_date = vulnerabilitiesimport['updated_date']
                        if updated_date == 'today':
                            updated_date = datetime.date.today()
                        log.debug(f"Updating vulnerabilities to {updated_date}")
                        log.debug(f"Vulnerabilities: {vulnerabilitylist}")
                        vulnerabilitylist = vulnerabilitylist.updated(updated_date)
                        log.debug(f"Updated vulnerabilities: {vulnerabilitylist}")
                    
                    log.debug(f"Setting vulnerabilities as {vulnerabilitylist}")
                    new_headers[vulnerabilitiesimport['variable']] = vulnerabilitylist

                    # Generate vulnerability markdown
                    vulnerabilitylist.generate_markdown(env)
                except FileNotFoundError as e:
                    log.warn(e.args[0])
                except Exception as e:
                    log.warn(f"Error loading vulnerabilities import {vulnerabilitiesimport['variable']}: {e.args}")
                    log.warn(traceback.format_exc())
            
        if 'sections' in headers['import']:
            sectionlist = make_list(headers['import']['sections'])
            for sectionimport in sectionlist:
                if not isinstance(sectionimport, dict):
                    log.warn(
                        "Section import item in front matter not a dict, skipping")
                    continue
                if 'directory' not in sectionimport or not isinstance(sectionimport['directory'], str):
                    log.warn(
                        "Section import item does not have a directory, skipping")
                    continue
                if 'variable' not in sectionimport or not isinstance(sectionimport['variable'], str):
                    log.warn(
                        "Section import item does not have variable name, skipping")
                    continue
                if sectionimport['variable'] in headers:
                    log.warn(
                        "Section import variable name already in front matter, skipping")
                    continue
                if 'ordinal' in sectionimport:
                    ordinal = sectionimport['ordinal']
                else:
                    ordinal = '1'

                # Build the section
                newsection = section.Section(
                    sectionimport['directory'], ordinal, watcher=watcher)

                new_headers[sectionimport['variable']] = newsection.get_chapters(env)

        if 'toml' in headers['import']:
            tomllist = make_list(headers['import']['toml'])
            if not hastomlib:
                log.warn("Cannot import TOML as we cannot import hastomlib")
            for tomlimport in tomllist:
                if not isinstance(tomlimport, str):
                    log.warn(
                        "TOML import item in front matter not a dict, skipping")
                    continue

                try:
                    with open(tomlimport, "rb") as f:
                        data = tomllib.load(f)
                        new_headers.update(data)
                        if watcher:
                            watcher.add_path(tomlimport)
                except Exception as e:
                    log.warn(f"Could not load toml file {tomlimport}: {e.args}")
        # Begin CSV imports
        if 'csv' in headers['import']:
            csvlist = make_list(headers['import']['csv'])
            for csvimport in csvlist:
                if not isinstance(csvimport, dict):
                    log.warn(
                        "csv import item in front matter not a dict, skipping")
                    continue
                if 'file' not in csvimport or not isinstance(csvimport['file'], str):
                    log.warn(
                        "csv import item does not have file location, skipping")
                    continue
                if 'variable' not in csvimport or not isinstance(csvimport['variable'], str):
                    log.warn(
                        "csv import item does not have variable name, skipping")
                    continue
                if csvimport['variable'] in headers:
                    log.warn(
                        "csv import variable name already in front matter, skipping")
                    continue
                if 'as_dict_list' in csvimport and not isinstance(csvimport['as_dict_list'], bool):
                    log.warning(
                        'csv import defines "as_dict_list", but it is not a boolean')
                    continue
                if 'index_col' in csvimport and not isinstance(csvimport['index_col'], int):
                    log.warning(
                        'csv import defines "index_col", but it is not an integer')
                    continue

                new_headers[csvimport['variable']] = import_csv(csvimport['file'], as_dict_list=csvimport.get(
                    'as_dict_list'), index_col=csvimport.get('index_col'))
                if watcher:
                    watcher.add_path(csvimport['file'])
        # Begin xlsx imports
        if 'xlsx' in headers['import']:
            xlsxlist = make_list(headers['import']['xlsx'])
            for xlsximport in xlsxlist:
                if not isinstance(xlsximport, dict):
                    log.warn(
                        "xlsx import item in front matter not a dict, skipping")
                    continue
                if 'file' not in xlsximport or not isinstance(xlsximport['file'], str):
                    log.warn(
                        "xlsx import item does not have file location, skipping")
                    continue
                if 'variable' not in xlsximport or not isinstance(xlsximport['variable'], str):
                    log.warn(
                        "xlsx import item does not have variable name, skipping")
                    continue
                if xlsximport['variable'] in headers:
                    log.warn(
                        "xlsx import variable name already in front matter, skipping")
                    continue
                if 'as_dict_list' in xlsximport and not isinstance(xlsximport['as_dict_list'], bool):
                    log.warning(
                        'xlsx import defines "as_dict_list", but it is not a boolean')
                    continue
                if 'index_col' in xlsximport and not isinstance(xlsximport['index_col'], int):
                    log.warning(
                        'xlsx import defines "index_col", but it is not an integer')
                    continue

                new_headers[xlsximport['variable']] = import_xlsx(
                    xlsximport['file'],
                    xlsximport.get('worksheet'),
                    xlsximport.get('min_row'),
                    xlsximport.get('max_row'),
                    xlsximport.get('min_col'),
                    xlsximport.get('max_col'),
                    as_dict_list=xlsximport.get('as_dict_list'),
                    index_col=xlsximport.get('index_col')
                )
                if watcher:
                    watcher.add_path(xlsximport['file'])

        if 'macros' in headers['import']:
            macrolist = make_list(headers['import']['macros'])
            for macrofile in macrolist:
                fn, ext = os.path.splitext(macrofile)
                if ext not in ['.md', '.j2', '.jinja', '.rr']:
                    log.warn(f"Macro file {macrofile} not of the right extension. The extension is {ext} but should be one of md, j2, jinja, or rr")
                    continue
                try:
                    macrofile = _get_resultant_path(macrofile, includemapper, pathlist)
                except:
                    log.warn(f"Could not find {macrofile}")
                    continue

                try:
                    mdheaders, markdown = markdown_from_file(macrofile, env=env, watcher=watcher)
                except:
                    log.warn(f"Could not read macro file {macrofile}")
                    continue
                mdheaders['cwd'] = os.path.dirname(
                    os.path.join(os.path.curdir, macrofile))

                try:
                    jenv = jinja2.Environment()
                    # Add in the env to the globals. We want the existing globals to take precedence
                    globals = env.get_env()
                    globals.update(jenv.globals)
                    jenv.globals = globals
                    # Get the macro template from the markdown
                    macro_template = jenv.from_string(markdown)
                    # Copy over the macros and set variables into the new headers
                    for f in [method for method in dir(macro_template.module) if method.startswith('_') is False]:
                        k = getattr(macro_template.module, f)
                        if isinstance(k, Macro):
                            log.debug(f"Setting macro {k.name}")
                            new_headers[k.name] = k
                        elif isinstance(k, str) or isinstance(k, int):
                            log.debug(f"Setting variable {f} as {k}")
                            new_headers[f] = k
                except jinja2.exceptions.TemplateSyntaxError as error:
                    log.error("Jinja2 error processing macro file {}: {} at lineno {} for file {}".format(
                        macrofile, error.message, error.lineno, filename=error.filename))
                    log.error("Skipping import of macro file {}".format(macrofile))
                    log_jinja2_error(markdown, error)
                    log.error(traceback.format_exc())
                    return ""
                except Exception as e:
                    log.error(
                        "Received exception when processing markdown for macro file {}: {}".format(macrofile, e.args))
                    log.error(
                        "Skipping import of macro file {}".format(macrofile))
                    log.error(traceback.format_exc())
                    return ""

    return new_headers


def markdown_from_file(file_loc, env=None, includemapper={}, process_includes=True, process_imports=True, pathlist=[], watcher=None):
    """ Read the markdown from a file.

    Arguments:
    - file_loc: The location of the file to read
    - env: The environment in its current context. The stack for this environment will be extended.
    - process_includes: Whether or not to process the include header, allowing the including of headers and markdown from other files.
    - process_imports: Whether or not to process the import header, allowing importing of sections, vulns, csv and xlsx files.
    - watcher: if a watcher is configured then add_path will automatically be called
    
    - watcher: if a watcher is configured then add_path will automatically be called
    
    Returns (headers, markdown) where headers is a dict and markdown is a long string """

    log.debug(f"Reading markdown from file {file_loc}")

    if not os.path.exists(file_loc):
        log.warn(f"Trying to read file {file_loc} but it doesn't exist")
        raise InputError("Could not read markdown file.")

    if watcher:
        watcher.add_path(file_loc)

    with open(file_loc, 'r') as vf:
        # If the first line starts with "---" that's our headers We have to make sure that it starts with "---"
        firstline = vf.readline()
        headers = dict()
        if yamlmatch.match(firstline):
            line = vf.readline()
            headerstring = ''
            # Read the headers.
            # Go until we get to the terminator to end the headers
            while not yamlmatch.match(line):
                headerstring += line
                line = vf.readline()
                if not line:
                    raise InputError(
                        f"Markdown file {file_loc} YAML headers don't end in a YAML terminator (...,---,===)")
            headers = yaml.safe_load(headerstring)  # Retrieve the YAML headers
            if headers == None:
                headers = {}
            markdown = vf.read()  # Read the rest of the file
        elif jsonmatch.match(firstline):
            line = vf.readline()
            headerstring = ''
            # Read the headers.
            # Go until we get to the terminator to end the headers
            while not jsonmatch.match(line):
                headerstring += line
                line = vf.readline()
                if not line:
                    raise InputError(
                        f"Markdown file {file_loc} JSON headers don't end in a JSON terminator (;;;)")
            headers = json.loads(headerstring)  # Retrieve the JSON headers
            markdown = vf.read()  # Read the rest of the file
        else:
            # There's no headers in this file. Put the first line back and just add it all into the markdown
            markdown = firstline + '\n' + vf.read()

        # Begin post processing. Process includes and imports.

        # Process includes in headers
        if process_includes and 'include' in headers:
            headers = _process_includes(headers, includemapper=includemapper, pathlist=pathlist, env=env)
        
        # Process include_markdown
        if process_includes and 'include_markdown' in headers:
            markdown = _process_include_markdown(headers['include_markdown'], includemapper, pathlist)

        # Process imports in headers
        if process_imports and 'import' in headers:
            if env == None:
                log.warn("Importing stopped because passed environment is None")
            else:
                log.info(f"Processing imports in file {file_loc}")
                headers.update(_process_imports(headers, env, watcher=watcher, pathlist=pathlist, includemapper=includemapper))

        return headers, markdown


def markdown_from_directory(dir_loc, env=None, watcher=None):
    log.debug("Reading markdown from directory '{}'".format(dir_loc))

    if not os.path.isdir(dir_loc):
        raise FileNotFoundError("Directory '{0}' doesn't exist. Skipping import of files from the directory.".format(
            dir_loc))

    files = []

    for f in os.listdir(dir_loc):
        mdfile = os.path.join(dir_loc, f)
        if os.path.isfile(mdfile):
            dir, filename = os.path.split(mdfile)
            fn, ext = os.path.splitext(filename)
            if ext == ".md" or ext == ".rr":
                try:
                    mdfromfile=markdown_from_file(mdfile, env, watcher=watcher)
                    mdfromfile[0]["filename"] = filename
                    mdfromfile[0]["filename_noext"] = fn
                    mdfromfile[0]["filename_full"] = mdfile
                    files.extend([mdfromfile])
                except:
                    log.warn(f"Could not read file {mdfile}")
                    continue
                

    return files



from report_ranger.output_formatter.outputformatter import OutputFormatter
import yaml
import html
from report_ranger.table import Table
import logging
import subprocess

log = logging.getLogger(__name__)

default_html_headers = {'template': ''}

html_header_aliases = {
}


class HTMLFormatter(OutputFormatter):
    def __init__(self, templateheaders=dict(), timer=None, watcher=None):
        OutputFormatter.__init__(self, templateheaders, timer, watcher=watcher)
        self.templatefile = self.templateheaders['html_template']
        self.figformat = "svg"

    def escape(self, text):
        ''' Escape the given text based on the format we're working with

        :param text: a plain text message
        :return: the message escaped to appear correctly in LaTeX
        '''
        return html.escape(text)

    def table(self, table, options={}, **tableargs):
        ''' This function formats the table in either latex or markdown, depending on the output. '''
        markdown = ""

        t = Table(table, env=self.env, **tableargs)

        markdown += "<table>"
        for i in range(len(t.table)):
            markdown += "\n<tr>"
            for j in range(len(t.table[i])):
                attrs = ""
                if t.colspan[i][j] < 1 or t.rowspan[i][j] < 1:
                    continue
                if t.colspan[i][j] > 1:
                    attrs += " colspan={}".format(t.colspan[i][j])
                if t.rowspan[i][j] > 1:
                    attrs += " rowspan={}".format(t.rowspan[i][j])
                if t.cellstyles[i][j] != '':
                    attrs += " class='{}'".format(t.cellstyles[i][j])

                # cell alignment, first try cellalign then colalign
                if t.cellalign[i][j] != '':
                    alignment = t.cellalign[i][j]
                elif t.colalign[j] != '':
                    alignment = t.colalign[j]
                else:
                    alignment = ''

                if alignment != '':
                    if alignment == 'l':
                        attrs += " style='text-align: left'"
                    elif alignment == 'c':
                        attrs += " style='text-align: center'"
                    elif alignment == 'r':
                        attrs += " style='text-align: right'"
                    elif alignment == 'j':
                        attrs += " style='text-align: justified'"

                markdown += "<td{}>{}</td>".format(attrs,
                                                   self.escape(str(t.table[i][j])))
            markdown += "</tr>"
        markdown += "</table>"
        return markdown

    def headers(self):
        markdown = ''
        headers = dict()
        headers['title'] = '{{title}}'
        headers['date'] = '{{date.strftime("%-d %B %Y")}}'

        # Put in the defaults for the latex template
        headers.update(default_html_headers)

        if "html" in self.template:
            for header in html_header_aliases.keys():
                if header in self.template['html']:
                    self.template['html'][html_header_aliases[header]
                                          ] = self.template['html'][header]
                    del self.template['html'][header]

            headers.update(self.template['html'])

        # Overwrite the headers with the report headers if they are set
        htmlheaders = self.env.get("html")
        if htmlheaders != None:
            for header in html_header_aliases.keys():
                if header in htmlheaders:
                    htmlheaders[html_header_aliases[header]
                                ] = htmlheaders[header]
                    del htmlheaders

            headers.update(htmlheaders)

        markdown += yaml.dump(headers)
        return markdown

    def output(self, markdown, outputfile=''):
        output = self._build_markdown(markdown)

        log.info("Writing HTML")

        # Use Pandoc to print to PDF
        pandoc_arguments = ['pandoc', '--from', 'markdown', '--to', 'html',
                            '--template', self.templatefile, '--listings', '-o', outputfile]
        log.info("Running pandoc with arguments {}".format(pandoc_arguments))
        process = subprocess.run(pandoc_arguments,
                                 input=output,
                                 universal_newlines=True)

        return output



from report_ranger.output_formatter.outputformatter import OutputFormatter
import subprocess
import logging

log = logging.getLogger(__name__)


class DOCXFormatter(OutputFormatter):
    def __init__(self, templateheaders=dict(), timer=None, watcher=None):
        OutputFormatter.__init__(self, templateheaders, timer, watcher=watcher)
        self.figformat = "png"

    def _build_markdown_headers(self, headers):
        return ""

    def output(self, markdown, outputfile=''):
        output = self._build_markdown(markdown)

        if(self.docformat and self.docformat != 'docx'):
            log.info("Writing Markdown")
            with open(outputfile, 'w') as fh:
                fh.write(output)
            log.info("Finished writing")

            return output

        log.info("Writing DOCX")
        # Use Pandoc to print to PDF
        pandoc_arguments = ['pandoc', '--from', 'markdown', '--to', 'docx',
                            '--listings', '-o', outputfile]
        if self.templatefile:
            pandoc_arguments.append('--template')
            pandoc_arguments.append(self.templatefile)
        process = subprocess.run(pandoc_arguments,
                                 input=output,
                                 universal_newlines=True)

        return output



from copy import deepcopy
import logging

log = logging.getLogger(__name__)


class Environment:
    """ Store environment variables

    This class operates as a stack. You can push a bunch of variables onto the stack when you go down a layer
    (say when a file is included). You can then pop off the stack when we finish with that file.
    """

    def __init__(self, other=None):
        self._static = {}
        self._variable = [{}]
        self._private = [{}]

        if other != None:
            # We can duplicate another environment. This involved deep copying so we don't accidentally
            # modify the other environment
            self._static = deepcopy(other._static)
            self._variable = deepcopy(other._variable)

    def set_private(self, key, value):
        # Set the private variable in the most recent stack
        # _private allows a set of hidden variables that can be kept track of through files
        # These variables cannot be accessed within the report
        self._private[-1][key] = value
    
    def get_private(self, key):
        # Go backwards through self._private looking to see if the variable is there
        for ve in reversed(self._private):
            if key in ve:
                return ve[key]


    def set_static(self, key, value):
        '''Set a static variable in the environment

        Static variables cannot be overwritten, they always win in a clash. This means they survive pushes and pops.
        '''
        self._static[key] = value

    def set_variable(self, key, value):
        '''Set a variable in the environment

        This variable will be set in the current layer of the stack. For instance, if pop is called the variable is lost.
        '''
        # When we set a variable we only set it to the last piece of the stack
        self._variable[-1][key] = value

    def set_variables(self, env=dict()):
        '''Set multiple variables in the environment

        These variables should be passed as a dict.
        '''
        self._variable[-1].update(env)

    def get(self, key):
        '''Get the variable with the key "key".'''
        # Static wins in a variable clash, so see if it's there first
        if key in self._static:
            return self._static[key]

        # Go backwards through self._variable looking to see if the variable is there
        for ve in reversed(self._variable):
            if key in ve:
                return ve[key]

        # We haven't found this variable
        return None

    def get_env(self):
        '''Get the current variables from the stack

        This function returns a dict full of key value pairs representing the current state of the environment.'''
        # We get the environment by overlaying each layer of the stack and then returning
        # what is at the end.
        env = {}
        for varenv in self._variable:
            env.update(varenv)

        # Static always wins, so finally overlay that
        env.update(self._static)
        return env

    def push(self, newenv={}):
        '''Add a layer to the stack

        This allows an isolated environment, where the variables you set can be removed once it's done.

        For instance, if you include a file, the headers of the included file will not affect the outer file.
        '''
        self._variable.append(newenv)
        self._private.append({})

    def pop(self):
        '''Take a layer off the stack

        This will remove all variables that have been added since the last time push() was called. Overwritten variables
        will be replaced with what was there before.
        '''
        self._variable.pop()
        self._private.pop()



import logging
from watchdog.events import FileSystemEventHandler
import pathlib
import time

log = logging.getLogger(__name__)

class Watcher(FileSystemEventHandler):
    def __init__(self, callback=None, *args, **kwargs):
        FileSystemEventHandler.__init__(self)
        self.paths = []
        self.callback = callback
        self.args = args
        self.kwargs = kwargs
        self.to_run = False
        self.os_run = False
        self.watch_mode = ''
        self.prevtime = time.time()
        self.running = False
    
    def set_callback(self, callback, *args, **kwargs):
        self.callback = callback
        self.args = args
        self.kwargs = kwargs
    
    def set_running(self):
        self.running = True

    def stop_running(self):
        self.running = False

    def set_watch_mode(self, watch_mode):
        self.watch_mode = watch_mode

    def add_path(self, path):
        self.paths.append(pathlib.Path(path))
        log.debug(f"Added {path} to watch paths")
    
    def clear_paths(self):
        self.paths.clear()

    def in_paths(self, path):
        for p in self.paths:
            if pathlib.Path(path).resolve() == p.resolve():
                return True
        return False

    def on_moved(self, event):
        log.debug(f"Detected move: {event.src_path}")
        if self.in_paths(event.dest_path) or self.in_paths(event.src_path):
            self.to_run = True
            self.os_run = True
    
    def on_modified(self, event):
        log.debug(f"Detected modification: {event.src_path}")
        path = pathlib.Path(event.src_path)
        if self.in_paths(event.src_path):
            if self.prevtime < path.stat().st_mtime or self.prevtime < path.stat().st_ctime:
                self.to_run = True
                self.os_run = True

    def on_deleted(self, event):
        log.debug(f"Detected deletion: {event.src_path}")
        if self.in_paths(event.src_path):
            self.to_run = True
            self.os_run = True

    def run(self):
        if self.running:
            log.debug("Still running, skipping recompile")
            return
        if self.watch_mode == "modified":
            for path in self.paths:
                if path.is_dir():
                    for df in [pth for pth in pathlib.Path(path).iterdir() if pth.suffix == '.md']:
                        mtime = df.stat().st_mtime
                        if self.prevtime < df.stat().st_mtime or self.prevtime < df.stat().st_ctime:
                            self.to_run = True
                            break
                elif path.is_file():
                    mtime = path.stat().st_mtime
                    if self.prevtime < mtime:
                        self.to_run = True
                        break
                else:
                    self.to_run = True
            
            self.prevtime = time.time()
        if self.to_run == True or self.os_run == True:
            log.info("Modifications detected, recompiling report.")
            self.callback(*self.args, **self.kwargs)
            self.to_run = False
            self.os_run = False
        else:
            log.debug(f"No modifications found")
    
    



import re
import logging
from report_ranger.utils.mdread import process_template
from report_ranger.helpers import filter_rows as fr


log = logging.getLogger(__name__)


def style_text_match(style_text, text, regexlist=None):
    ''' See if the text is in the style_text dict and match it to stylelist '''
    if type(text) is not str:
        return None
    if not regexlist:
        regexlist = [(re.compile(ts), cf)
                     for (ts, cf) in style_text.items()]

    # We have to go through each regex to see if it matches
    for ts in regexlist:
        # Regexes are in a tuple (regex, cf)
        if ts[0].match(text):
            return ts[1]

def convert_list_of_dicts_to_matrix(table):
    tableheadings = []

    # Get the headings from the dicts
    for row in table:
        for h in row.keys():
            if h not in tableheadings:
                tableheadings.append(h)

    newtable = []

    # Headings will always be the first row of the new table
    newtable.append(tableheadings)
    for row in table:
        newrow = []
        for h in tableheadings:
            if h in row.keys():
                newrow.append(row[h])
            else:
                # This heading isn't in this row, add a blank
                newrow.append("")
        newtable.append(newrow)
    return newtable

def process_list_of_dicts(table, append_column, env):
    """Convert a list of dicts to a matrix with a header line
    
    append_column will be appended to the end of each row.
    """
    
    tableheadings = None

    for r in range(len(table)):
        if not isinstance(table[r], dict):
            log.warn(
                f"Table defined as dicts has something which is not a dict, replacing with empty row: {table[r]}")
            table[r] = dict()

        # Process append_column
        if isinstance(append_column, dict):
            append_column_done = True
            newcols = {}
            for key in append_column:
                newcols[key] = process_template(
                    table[r], append_column[key], env=env, name="append_column")
            table[r].update(newcols)

    tableheadings = []

    # Get the headings from the dicts
    for row in table:
        for h in row.keys():
            if h not in tableheadings:
                tableheadings.append(h)

    newtable = []

    # Headings will always be the first row of the new table
    newtable.append(tableheadings)
    for row in table:
        newrow = []
        for h in tableheadings:
            if h in row.keys():
                newrow.append(row[h])
            else:
                # This heading isn't in this row, add a blank
                newrow.append("")
        newtable.append(newrow)
    return newtable, tableheadings


class Table:
    """ This class holds a table to be used in output formatters. It will hold things like table contents, headings, alignment, widths

    The table has the following state:
    table: A list of lists containing the content of each cell, always guaranteed to be square. Table will pad if necessary.
    cellstyles: A list of lists containing a string with the style of each square, always guaranteed be the same size of table.
    colalign: A list of strings, either 'l', 'c', 'r', 'j' or an empty string. Guaranteed to be the width of the table.
    cellalign: A list of lists of strings,  either 'l', 'c', 'r', 'j' or an empty string, always guaranteed be the same size of table.
    colspan: A list of lists containing numbers where each number >= 1 or -1 if the cell is overwritten, always guaranteed be the same size of table. This is for when you would like a cell to overwrite the next column.
    rowspan: A list of lists containing numbers where each number >= 1 or -1 if the cell is overwritten, always guaranteed be the same size of table. This is for when you would like a cell to overwrite the next row.
    colwidths: A single list of numbers. Guaranteed to be the width of the table. This corresponds to the width of each cell.
    """

    def _pad_matrix(self, m, width=0, height=0, pad_value=''):
        """ Pad the matrix m (a list of lists) to the set width and height with the pad_value.

        If the width and height is not set, make sure that the matrix m is rectangular, with each row being the same length by padding them with pad_value.
        """
        if m == None:
            m = []
            for i in range(height):
                m += [[pad_value]*width]
            return m

        if height == 0:
            height = len(m)

        if width == 0:
            for i in m:
                mw = len(i)
                if mw > width:
                    width = mw

        for row in m:
            if len(row) < width:
                row += [pad_value]*(width-len(row))
            if len(row) > width:
                row = row[:width]

        if len(m) < height:
            for i in range(height-len(m)):
                m += [[pad_value]*width]
        elif len(m) > height:
            m = m[:height]

        return m

    def _pad_list(self, l, width, pad_value=''):
        """ Pad the list l to the required width with pad_value """
        if l == None:
            l = []

        if len(l) < width:
            l += [pad_value]*(width-len(l))

        return l
    
    def _clear_table(self):
        self.table = [[]]
        self.width = 0
        self.height = 0
        self.cellstyles = []
        self.colalign = []
        self.cellalign = []
        self.colspan = []
        self.rowspan = []
        self.colwidths = []
    
    def _markdown_to_matrix(self, markdown):
        """ Convert the markdown string into a table.
        
        Returns """
        re_tablecells = re.compile(r'(\||^)([^\|]*)(?=(\||$))')
        re_isheaderline = re.compile(r'^\|?(\s*:?-+:?\s*\|?)*\s*$')
        re_hastablecells = re.compile(r'\|')
        lines = markdown.splitlines()
        matrixtable = []
        colalign = []
        has_heading = False
        for line in lines:
            if not re_hastablecells.match(line):
                continue
            cells = re_tablecells.findall(line)
            # Is it the heading line?
            if re_isheaderline.match(line):
                has_heading = True
                for cell in cells:
                    ic = cell[1].strip()
                    if len(ic) == 0:
                        continue
                    if ic[0] == ':':
                        if ic[-1] == ':':
                            colalign.append('c')
                        else:
                            colalign.append('l')
                    elif ic[-1] == ':':
                        colalign.append('r')
                    else:
                        colalign.append('l')
            else:
                row = []
                # Check to see if the first line has a |, if so skip the last one
                if cells[0][0] == '|':
                    endpipe = True
                for cell in cells:
                    row.append(cell[1].strip())
                if endpipe:  # Kill last empty cell
                    if row[-1] == '':
                        row = row[:-1]
                matrixtable.append(row)

        return matrixtable, colalign, has_heading

    
        

    def __init__(self, table, env=None, filter_rows=None, header=[], headings=None, cellstyles=None, colstyles=None, colalign=None, cellalign=None, colspan=None, rowspan=None, colwidths=None, append_column=None, colpicker=None, rowpicker=None, style_text={}):
        # Is it just a markdown table?
        if isinstance(table, str):
            table, newcolalign, has_heading = self._markdown_to_matrix(table)

            # Append colalign if it's not complete
            if colalign == None:
                colalign = []
            colalign += newcolalign[len(colalign):]
            
            # Set a heading row if it has one
            if has_heading == True and headings == None:
                headings = 'top'

        # The table now has to be a list. If it is not a list, warn and skip
        if not isinstance(table, list):
            log.warn(f"Table not a list. Trying to make a table out of {table}.")
            self._clear_table()
            return
        
        # If there's no rows there's nothing to display
        if len(table) == 0:
            log.warn("Table has no rows, skipping table.")
            self._clear_table()
            return

        # We perform validation now, after converting a blank markdown table and before converting list of dicts to list of lists
        if filter_rows:
            table = fr(table, filter_rows)
            if len(table) == 0:
                log.warning("Table has no rows after filtering")
                self._clear_table()
                return
            
        append_column_done = False

        # We are representing the table in a dict form and we need to translate that into a table. For instance:
        # affected_hosts:
        # - hostname: host.com
        #   port: 80
        # - hostname: host2.com
        #   port: 80
        tableheadings = None
        if isinstance(table[0], dict):
            table, tableheadings = process_list_of_dicts(table, append_column, env)
            append_column_done = True

        # We should now have a list of lists. Let's just make sure!
        newtable = []
        for row in table:
            if not isinstance(row, list):
                log.warn(f"Table row not a list. Skipping table row: {row}")
            else:
                newtable.append(row)
        table = newtable

        # If we have a list of lists, then append_column won't be done yet. Let's do that.
        if append_column and not append_column_done:
            if isinstance(append_column, list):
                for row in range(len(table)):
                    rowdict = {}
                    # We need to index for each column
                    for col in range(len(row)):
                        rowdict['col' + str(col)] = table[row][col]
                    for key in range(len(append_column)):
                        table[row].append(process_template(
                            rowdict, append_column[key], env=env, name="append_column"))
            elif isinstance(append_column, dict):
                log.warning(f"append_column a dict for a table of lists, skipping: {append_column}")
            else:
                log.warning(f"append_column not a list, skipping: {append_column}")

        # Pad everything
        self.table = self._pad_matrix(table, pad_value='')
        width, height = len(table[0]), len(table)
        
        self.width, self.height = width, height

        # Handle rowpicker and colpicker
        # Rowpicker first, since filtering rows is easier and quicker than filtering columns
        if rowpicker != None and rowpicker != []:
            if not isinstance(rowpicker, list):
                log.warn("Rowpicker variable not a list, ignoring.")
            else:
                newtable = []
                for row in rowpicker:
                    if not isinstance(row, int):
                        log.warn(
                            "Entry in rowpicker was not int. Found: {}".format(row))
                    elif row < 0 or row >= len(self.table):
                        log.warn("Entry in rowpicker outside the range of the table length {}. Found: {}".format(
                            len(self.table), row))
                    else:
                        newtable.append(self.table[row])
                self.table = newtable
                self.height = len(self.table)

        if colpicker != None and colpicker != []:
            if not isinstance(colpicker, list):
                log.warn("Colpicker variable not a list, ignoring.")
            else:
                newtable = []
                for i in range(len(self.table)):
                    newtable.append([])
                for col in colpicker:
                    if isinstance(col, str):
                        # Check to see if we're referring to a column heading
                        if col in self.table[0]:
                            col = self.table[0].index(col)
                        else:
                            log.warn(
                                "Entry in colpicker is a string but not a column heading. Found: {}".format(col))
                            continue
                    elif not isinstance(col, int):
                        log.warn(
                            "Entry in colpicker was not int. Found: {}".format(col))
                        continue

                    if col < 0 or col >= len(self.table[0]):
                        log.warn("Entry in colpicker outside the range of the table length {}. Found: {}".format(
                            len(self.table[0]), col))
                        continue

                    for trow in range(len(self.table)):
                        newtable[trow].append(self.table[trow][col])
                self.table = newtable

                if len(self.table) > 0:  # Do we actually still have a table?
                    self.width = len(self.table[0])
                else:
                    self.width = 0

        # If there's no table anymore due to rowpicker or colpicker then get rid of the rest, just cancel it out
        if self.width == 0:
            log.warn("rowpicker and colpicker resulted in an empty table.")
            self._clear_table()
            return

        # Put in the header if it's been supplied
        if header != []:
            if isinstance(header, list):
                # If there's tableheadings from a dict already, remove them
                if tableheadings:
                    self.table = [header] + self.table[1:]

                else:
                    self.table = [header] + self.table

                # Do we need to repad?
                if width != len(header):
                    self.table = self._pad_matrix(self.table, pad_value='')
                    self.width = len(self.table[0])

                self.height = len(self.table)
            else:
                log.warn(
                    "Header of the table is not a list. Trying to add {}.".format(header))

        # colalign must be one of 'l' 'c' 'r' or 'j' or an empty string
        colalign = self._pad_list(colalign, self.width, '')
        for i in range(self.width):
            if colalign[i] not in 'lcrj':
                colalign[i] = ''

        cellalign = self._pad_matrix(cellalign, self.width, self.height, '')
        for i in range(self.height):
            for j in range(self.width):
                if cellalign[i][j] not in 'lcrj':
                    cellalign[i][j] = ''
        self.colalign = colalign
        self.cellalign = cellalign

        # Handle colspan and rowspan
        colspan = self._pad_matrix(
            colspan, self.width, self.height, 1)  # Pad to size of table
        rowspan = self._pad_matrix(
            rowspan, self.width, self.height, 1)  # Pad to size of table

        # Handle colspan
        for i in range(self.height):
            for j in range(self.width):
                if not int(colspan[i][j]):  # Validate it's an int
                    colspan[i][j] = 1
                if colspan[i][j] >= 1:  # We have a colspan!
                    # Does it go over the side of the table?
                    if j + colspan[i][j] > self.width:
                        colspan[i][j] = self.width - j  # Snip it off
                    # We blank out the rest of the span
                    for span in range(1, colspan[i][j]):
                        colspan[i][j+span] = -1

        # Handle rowspan. This is equivalent of the above, just swapping row and column
        for i in range(self.height):
            for j in range(self.width):
                if not int(rowspan[i][j]):
                    rowspan[i][j] = 1
                if rowspan[i][j] >= 1:
                    if i + rowspan[i][j] > self.height:
                        rowspan[i][j] = self.height - i
                    for span in range(1, rowspan[i][j]):
                        rowspan[i+span][j] = -1
        self.colspan = colspan
        self.rowspan = rowspan

        self.colwidths = self._pad_list(colwidths, self.width, 0)

        cellstyles = self._pad_matrix(cellstyles, self.width, self.height, '')

        # Allow the 'left', 'top', and 'left-top' headings settings
        if not isinstance(headings, list) and not headings == None:
            headingslist = headings.split('-')
            if 'left' in headingslist:
                for row in cellstyles:
                    if row[0] == '':
                        row[0] = 'h'
            if 'right' in headingslist:
                for row in cellstyles:
                    if row[-1] == '':
                        row[-1] = 'h'
            if 'top' in headingslist:
                for i in range(len(cellstyles[0])):
                    if cellstyles[0][i] == '':
                        cellstyles[0][i] = 'h'
            if 'bottom' in headingslist:
                for i in range(len(cellstyles[-1])):
                    if cellstyles[-1][i] == '':
                        cellstyles[-1][i] = 'h'
        else:
            cellstyles = self._pad_matrix(
                cellstyles, self.width, self.height, '')

        # Do the col styles if set. Go through and set each entire column in cellstyles as per what is set by the user in colstyles.
        # Note that this won't overwrite headers
        if colstyles:
            for row in range(len(cellstyles)):
                for col in range(len(cellstyles[row])):
                    # Don't overwrite existing cellstyles, only fill in if it's blank
                    if (cellstyles[row][col] == '' or cellstyles[row][col] == None) and col < len(colstyles) and colstyles[col] and colstyles[col] != '':
                        cellstyles[row][col] = colstyles[col]

        # If there's headings text sync then add them
        if style_text:
            # Get all the regexes
            stregexlist = [(re.compile(ts), cf)
                           for (ts, cf) in style_text.items()]

        for r in range(self.height):
            for c in range(self.width):
                # First check to see if there's a style in cellstyles
                if type(cellstyles[r][c]) is dict:
                    # If it's a dict, treat it as a style text dict
                    st_return = style_text_match(
                        cellstyles[r][c], self.table[r][c])
                    if st_return:
                        cellstyles[r][c] = st_return
                    else:
                        cellstyles[r][c] = ''
                elif type(cellstyles[r][c]) is str and cellstyles[r][c] != '':
                    # Leave it as is if there's something filled in
                    continue
                elif style_text:
                    # Now check master style_text
                    cellstyle = style_text_match(
                        style_text, self.table[r][c], stregexlist)
                    if cellstyle:
                        cellstyles[r][c] = cellstyle
                else:
                    cellstyles[r][c] = ''

        self.cellstyles = cellstyles



# This file contains validation functions and exceptions

import cerberus
import logging

log = logging.getLogger(__name__)

default_report_validation = {
    "title": {'type': 'string', 'required': True}
}

default_vulnerability_validation = {
    "name": {'type': 'string', 'required': True}
}

default_appendix_validation = {
    "name": {'type': 'string', 'required': True}
}


class ValidationError(Exception):
    pass


def validate_headers(validation, headers, defaults={}, name="template"):
    validator = cerberus.Validator()
    validator.allow_unknown = True

    schema = defaults.copy()
    schema.update(validation)

    validator.schema = schema

    if validator.validate(headers) == True:
        return True

    # We have an error!
    for error in validator.errors.keys():
        logging.error("Validation of {} failed due to header {}: {}".format(name,
                                                                            error, validator.errors[error]))

    raise ValidationError(
        "Validation of headers of {} failed: {}".format(name, validator.errors))
